package dev.fitko.fitconnect.core.schema;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import dev.fitko.fitconnect.api.config.defaults.SchemaConfig;
import dev.fitko.fitconnect.api.domain.schema.SchemaResources;
import dev.fitko.fitconnect.api.exceptions.internal.SchemaNotFoundException;
import dev.fitko.fitconnect.api.services.schema.SchemaProvider;
import dev.fitko.fitconnect.core.http.HttpClient;
import dev.fitko.fitconnect.core.http.HttpResponse;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.io.TempDir;

import java.io.IOException;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.equalTo;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class SchemaResourceProviderTest {

    private SchemaProvider underTest;

    @BeforeEach
    void setup(@TempDir final Path tempDir) throws IOException {

        // prepare local test submission data schema
        final String schemaResource = getResourceAsString("/submission-data-test-schema/s00000096_1.0.schema.json");
        final Path submissionDataSchemaPath = Files.writeString(Path.of(tempDir.toString(), "submission_data_schema.json"), schemaResource);

        final List<String> setSchemas = SchemaConfig.getSetSchemaFilePaths("/set-schema");
        final List<String> metadataSchemas = SchemaConfig.getMetadataSchemaFileNames("/metadata-schema");
        final List<String> destinationSchemas = SchemaConfig.getDestinationSchemaPaths("/destination-schema");
        final SchemaResources schemaResources = new SchemaResources(setSchemas, metadataSchemas, destinationSchemas, submissionDataSchemaPath.toString());
        underTest = new SchemaResourceProvider(mock(HttpClient.class), schemaResources);
    }

    @Test
    void isAllowedSetSchema() {
        assertTrue(underTest.isAllowedSetSchema(SchemaConfig.SET_V_1_0_1.getSchemaUri()));
        assertTrue(underTest.isAllowedSetSchema(SchemaConfig.SET_V_1_0_0.getSchemaUri()));
        assertTrue(underTest.isAllowedSetSchema(URI.create("https://schema.fitko.de/fit-connect/set-payload/1.2.3/set-payload.schema.json")));
    }

    @Test
    void isNoAllowedSetSchema() {
        assertFalse(underTest.isAllowedSetSchema(URI.create("https://schema.fitko.de/fit-connect/set-payload/2.2.3/set-payload.schema.json")));
    }

    @Test
    void isAllowedMetadataSchema() {
        assertTrue(underTest.isAllowedMetadataSchema(SchemaConfig.METADATA_V_1_0_0.getSchemaUri()));
    }

    @Test
    void isNoAllowedMetadataSchema() {
        assertFalse(underTest.isAllowedSetSchema(URI.create("https://schema.fitko.de/fit-connect/metadata/9.9.9/metadata.schema.json")));
    }

    @Test
    void loadLatestSetSchema() throws JsonProcessingException {
        // Given
        final URI schemaUri = SchemaConfig.SET_V_1_1_0.getSchemaUri();

        // When
        final String setSchema = underTest.loadLatestSetSchema();

        //Then
        assertThat(new ObjectMapper().readTree(setSchema).get("$id").asText(), equalTo(schemaUri.toString()));
    }

    @Test
    void loadMetadataSchema() throws JsonProcessingException {
        // Given
        final URI schemaUri = SchemaConfig.METADATA_V_1_0_0.getSchemaUri();

        // When
        final String metadataSchema = underTest.loadMetadataSchema(schemaUri);

        //Then
        assertThat(new ObjectMapper().readTree(metadataSchema).get("$id").asText(), equalTo(schemaUri.toString()));
    }

    @Test
    void loadDestinationSchema() throws JsonProcessingException {
        // Given
        final URI schemaUri = SchemaConfig.XZUFI_DESTINATION_SCHEMA.getSchemaUri();

        // When
        final String metadataSchema = underTest.loadDestinationSchema(schemaUri);

        //Then
        assertThat(new ObjectMapper().readTree(metadataSchema).get("$id").asText(), equalTo(schemaUri.toString()));
    }

    @Test
    void loadMetadataSchemaThatDoesNotExist() {
        final URI schemaUri = URI.create("https://schema.fitko.de/fit-connect/metadata/9.9.9/metadata.schema.json");
        assertThrows(SchemaNotFoundException.class, () -> underTest.loadMetadataSchema(schemaUri));
    }

    @Test
    void loadSubmissionDataSchemaFromMemoryWorks() {

        final String schema = underTest.loadSubmissionDataSchema(URI.create("https://schema.fitko.de/fim/s00000096_1.0.schema.json"));

        assertThat(schema, containsString("\"$id\": \"https://schema.fitko.de/fim/s00000096_1.0.schema.json\""));
    }

    @Test
    void loadSubmissionDataSchemaFromMemoryFailsButRemoteWorks() {

        final HttpClient httpClient = mock(HttpClient.class);
        when(httpClient.get(eq("https://test.json"), any())).thenReturn(new HttpResponse<>(200, "schema"));

        final SchemaResourceProvider schemaResourceProvider = new SchemaResourceProvider(httpClient,
                new SchemaResources(List.of(), List.of(), List.of(), null));

        final String schema = schemaResourceProvider.loadSubmissionDataSchema(URI.create("https://test.json"));

        assertThat(schema, equalTo("schema"));
    }

    @Test
    void loadSubmissionDataSchemaFromMemoryAndRemoteFails() {

        final HttpClient httpClient = mock(HttpClient.class);
        when(httpClient.get(eq("https://schema.fitko.de/fim/s00000096_1.0.schema.json"),
                any())).thenThrow(new RuntimeException("Something went wrong!"));

        final SchemaResourceProvider schemaResourceProvider = new SchemaResourceProvider(httpClient,
                new SchemaResources(List.of(), List.of(), List.of(), null));

        final Exception exception = assertThrows(SchemaNotFoundException.class,
                () -> schemaResourceProvider.loadSubmissionDataSchema(URI.create("https://schema.fitko.de/fim/s00000096_1.0.schema.json")));

        assertThat(exception.getMessage(), equalTo("Submission data schema https://schema.fitko.de/fim/s00000096_1.0.schema.json is not available."));
    }

    @Test
    void loadSubmissionDataSchemaFromRemoteIsRejectedBecauseNoHttps() {

        final SchemaResourceProvider schemaResourceProvider = new SchemaResourceProvider(mock(HttpClient.class),
                new SchemaResources(List.of(), List.of(), List.of(), null));

        final Exception exception = assertThrows(SchemaNotFoundException.class,
                () -> schemaResourceProvider.loadSubmissionDataSchema(URI.create("http://test.json")));

        assertThat(exception.getMessage(), equalTo("Fetching schema http://test.json from remote was skipped, since the URI does not support HTTPS."));
    }

    String getResourceAsString(final String filename) throws IOException {
        return new String(SchemaResourceProviderTest.class.getResourceAsStream(filename).readAllBytes());
    }
}